#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <process.h>
#include <i86.h>
#include <sys/sched.h>
#include <sys/kernel.h>

#define false   0
#define true    !false

#define SLICE_MS        1
#define SLICE_TASKS     4
#define SLICE_PRIO      15
#define SLICE_RUN       5
#define SLICE_SLEEP     4

struct amsg
{
  msg_t type;
  msg_t status;
  uchar_t buf[128];
};

pid_t slice_pid[ SLICE_TASKS ];
volatile int slice_cur = SLICE_TASKS - 1;
char slice_name[ SLICE_TASKS ][80] = { "one",
                                       "two",
                                       "three",
                                       "four" };
struct amsg msg;
pid_t msgpid;
pid_t sched_proxy;

/* interrupt handler must be in separate file due to different compile flags */
extern pid_t far sched_clk( void );

/******************************************************************************

        MAIN (sigs)

******************************************************************************/
main( int argc, char *argv[] )
{
int ii;
int timer_id;
int timer_events;

  /* change timer ticksize to 1 ms */
  qnx_ticksize( 1000000 * SLICE_MS, _TICKSIZE_STANDARD );

  printf( "tick changed\n" );

  /* change to round robin - spawned tasks will inherit */
  setscheduler( 0, SCHED_RR, 9 );

  /* start the tasks in the list */
  for ( ii = 0; ii < SLICE_TASKS; ii++ )
  {
    /* spawn the tasks at current priority */
    /* note - each task should wait for its priority to be SLICE_RUN 
       before starting its processing */
    slice_pid[ii] = spawnl( P_NOWAIT, slice_name[ii], slice_name[ii], NULL );
    printf( "slice_name %s started\n", slice_name[ii] );
    if ( slice_pid[ii] == -1 )
      exit( -1 );
  }

  /* raise the priority of this task so we get control */
  setscheduler( 0, SCHED_RR, SLICE_PRIO );

  /* alter priorities of tasks - they will not run since below this code */
  for ( ii = 0; ii < SLICE_TASKS; ii++ )
  {
    if ( ii != slice_cur )
      setprio( slice_pid[ii], SLICE_SLEEP );
    else
      setprio( slice_pid[ii], SLICE_RUN );
    printf( "setprio for %d\n", slice_pid[ii] );
  }

  /* get a proxy id so we can perform scheduling when informed by timer */
  if ( (sched_proxy = qnx_proxy_attach( 0, NULL, 0, -1 ) ) == -1 )
    exit( -1 );

  /* attach to interrupt - begin managing time slices */
  if ( (timer_id = qnx_hint_attach( 0,  /* hardware timer */
                   sched_clk,
                   FP_SEG(slice_pid)) ) == -1 )
    exit( -1 );
  printf( "timer is attached\n" );

  while ( true )
  {
    /* wait 1 ms for a message from a process */
    msgpid = Receive( 0, &msg, sizeof(msg) );
    if ( msgpid == sched_proxy )
    {
      setprio( slice_pid[ slice_cur ], SLICE_SLEEP );
      slice_cur = ++slice_cur % SLICE_TASKS;
#if NORMAL
      setprio( slice_pid[ slice_cur ], SLICE_RUN );
#else
      /* this line shows "spreads" the time across SLICE_TASKS priorities
         shows up on photon phsac */
      setprio( slice_pid[ slice_cur ], SLICE_RUN + slice_cur );
#endif
      /* the following verifies that the scheduler is being run */
#if CHECK
      if ( !(++timer_events % 1000) )
        printf( "%d\n", timer_events );
#endif
    }
    else
    {
      msg.status = EOK;
      Reply( msgpid, &msg, sizeof(msg.status) );
    }
  }
}  /* main */

